export default makeE;
export type EProxy = ReturnType<(HandledPromise: HandledPromiseConstructor) => (<T>(x: T) => ECallableOrMethods<RemoteFunctions<T>>) & {
    /**
     * E.get(x) returns a proxy on which you can get arbitrary properties.
     * Each of these properties returns a promise for the property.  The promise
     * value will be the property fetched from whatever 'x' designates (or
     * resolves to) in a future turn, not this one.
     *
     * @template T
     * @param {T} x target for property get
     * @returns {EGetters<LocalRecord<T>>} property get proxy
     * @readonly
     */
    readonly get: <T>(x: T) => EGetters<LocalRecord<T>>;
    /**
     * E.resolve(x) converts x to a handled promise. It is
     * shorthand for HandledPromise.resolve(x)
     *
     * @template T
     * @param {T} x value to convert to a handled promise
     * @returns {Promise<Awaited<T>>} handled promise for x
     * @readonly
     */
    readonly resolve: {
        (): Promise<void>;
        <T>(value: T): Promise<Awaited<T>>;
        <T>(value: T | PromiseLike<T>): Promise<Awaited<T>>;
    };
    /**
     * E.sendOnly returns a proxy similar to E, but for which the results
     * are ignored (undefined is returned).
     *
     * @template T
     * @param {T} x target for method/function call
     * @returns {ESendOnlyCallableOrMethods<RemoteFunctions<T>>} method/function call proxy
     * @readonly
     */
    readonly sendOnly: <T>(x: T) => ESendOnlyCallableOrMethods<RemoteFunctions<T>>;
    /**
     * E.when(x, res, rej) is equivalent to
     * HandledPromise.resolve(x).then(res, rej)
     *
     * @template T
     * @template [U = T]
     * @param {T|PromiseLike<T>} x value to convert to a handled promise
     * @param {(value: T) => ERef<U>} [onfulfilled]
     * @param {(reason: any) => ERef<U>} [onrejected]
     * @returns {Promise<U>}
     * @readonly
     */
    readonly when: <T, U = T>(x: T | PromiseLike<T>, onfulfilled?: ((value: T) => ERef<U>) | undefined, onrejected?: ((reason: any) => ERef<U>) | undefined) => Promise<U>;
}>;
/**
 * Declare an object that is potentially a far reference of type Primary whose
 * auxilliary data has type Local.  This should be used only for consumers of
 * Far objects in arguments and declarations; the only creators of Far objects
 * are distributed object creator components like the `Far` or `Remotable`
 * functions.
 */
export type FarRef<Primary, Local = DataOnly<Primary>> = ERef<Local & import("./types.js").RemotableBrand<Local, Primary>>;
/**
 * `DataOnly<T>` means to return a record type `T2` consisting only of
 * properties that are *not* functions.
 */
export type DataOnly<T> = Omit<T, FilteredKeys<T, import("./types.js").Callable>>;
/**
 * Declare that `T` may or may not be a Promise.  This should be used only for
 * consumers of arguments and declarations; return values should specifically be
 * `Promise<T>` or `T` itself.
 */
export type ERef<T> = PromiseLike<T> | T;
/**
 * The awaited return type of a function.
 */
export type EReturn<T extends (...args: any[]) => any> = T extends (...args: any[]) => infer R ? Awaited<R> : never;
export type ECallable<T extends import("./types.js").Callable> = (ReturnType<T> extends PromiseLike<infer U> ? T : (...args: Parameters<T>) => Promise<EReturn<T>>);
export type EMethods<T> = { readonly [P in keyof T]: T[P] extends import("./types.js").Callable ? ECallable<T[P]> : never; };
export type EGetters<T> = { readonly [P in keyof T]: T[P] extends PromiseLike<infer U> ? T[P] : Promise<Awaited<T[P]>>; };
export type ESendOnlyCallable<T extends import("./types.js").Callable> = (...args: Parameters<T>) => Promise<void>;
export type ESendOnlyMethods<T> = { readonly [P in keyof T]: T[P] extends import("./types.js").Callable ? ESendOnlyCallable<T[P]> : never; };
export type ESendOnlyCallableOrMethods<T> = (T extends import("./types.js").Callable ? ESendOnlyCallable<T> & ESendOnlyMethods<Required<T>> : ESendOnlyMethods<Required<T>>);
export type ECallableOrMethods<T> = (T extends import("./types.js").Callable ? ECallable<T> & EMethods<Required<T>> : EMethods<Required<T>>);
/**
 * Return a union of property names/symbols/numbers P for which the record element T[P]'s type extends U.
 *
 * Given const x = { a: 123, b: 'hello', c: 42, 49: () => {}, 53: 67 },
 *
 * FilteredKeys<typeof x, number> is the type 'a' | 'c' | 53.
 * FilteredKeys<typeof x, string> is the type 'b'.
 * FilteredKeys<typeof x, 42 | 67> is the type 'c' | 53.
 * FilteredKeys<typeof x, boolean> is the type never.
 */
export type FilteredKeys<T, U> = { [P in keyof T]: T[P] extends U ? P : never; }[keyof T];
/**
 * `PickCallable<T>` means to return a single root callable or a record type
 * consisting only of properties that are functions.
 */
export type PickCallable<T> = (T extends import("./types.js").Callable ? (...args: Parameters<T>) => ReturnType<T> : Pick<T, FilteredKeys<T, import("./types.js").Callable>>);
/**
 * `RemoteFunctions<T>` means to return the functions and properties that are remotely callable.
 */
export type RemoteFunctions<T> = (T extends import("./types.js").RemotableBrand<infer L, infer R> ? PickCallable<R> : Awaited<T> extends import("./types.js").RemotableBrand<infer L, infer R> ? PickCallable<R> : T extends PromiseLike<infer U> ? Awaited<T> : T);
export type LocalRecord<T> = (T extends import("./types.js").RemotableBrand<infer L, infer R> ? L : Awaited<T> extends import("./types.js").RemotableBrand<infer L, infer R> ? L : T extends PromiseLike<infer U> ? Awaited<T> : T);
export type EPromiseKit<R = unknown> = {
    promise: Promise<R>;
    settler: import("./types.js").Settler<R>;
};
/**
 * Declare a near object that must only be invoked with E, even locally.  It
 * supports the `T` interface but additionally permits `T`'s methods to return
 * `PromiseLike`s even if `T` declares them as only synchronous.
 */
export type EOnly<T> = (T extends import("./types.js").Callable ? (...args: Parameters<T>) => ERef<Awaited<EOnly<ReturnType<T>>>> : T extends Record<PropertyKey, import("./types.js").Callable> ? { [K in keyof T]: T[K] extends import("./types.js").Callable ? (...args: Parameters<T[K]>) => ERef<Awaited<EOnly<ReturnType<T[K]>>>> : T[K]; } : T);
/**
 * @param {HandledPromiseConstructor} HandledPromise
 */
declare function makeE(HandledPromise: HandledPromiseConstructor): (<T>(x: T) => ECallableOrMethods<RemoteFunctions<T>>) & {
    /**
     * E.get(x) returns a proxy on which you can get arbitrary properties.
     * Each of these properties returns a promise for the property.  The promise
     * value will be the property fetched from whatever 'x' designates (or
     * resolves to) in a future turn, not this one.
     *
     * @template T
     * @param {T} x target for property get
     * @returns {EGetters<LocalRecord<T>>} property get proxy
     * @readonly
     */
    readonly get: <T>(x: T) => EGetters<LocalRecord<T>>;
    /**
     * E.resolve(x) converts x to a handled promise. It is
     * shorthand for HandledPromise.resolve(x)
     *
     * @template T
     * @param {T} x value to convert to a handled promise
     * @returns {Promise<Awaited<T>>} handled promise for x
     * @readonly
     */
    readonly resolve: {
        (): Promise<void>;
        <T>(value: T): Promise<Awaited<T>>;
        <T>(value: T | PromiseLike<T>): Promise<Awaited<T>>;
    };
    /**
     * E.sendOnly returns a proxy similar to E, but for which the results
     * are ignored (undefined is returned).
     *
     * @template T
     * @param {T} x target for method/function call
     * @returns {ESendOnlyCallableOrMethods<RemoteFunctions<T>>} method/function call proxy
     * @readonly
     */
    readonly sendOnly: <T>(x: T) => ESendOnlyCallableOrMethods<RemoteFunctions<T>>;
    /**
     * E.when(x, res, rej) is equivalent to
     * HandledPromise.resolve(x).then(res, rej)
     *
     * @template T
     * @template [U = T]
     * @param {T|PromiseLike<T>} x value to convert to a handled promise
     * @param {(value: T) => ERef<U>} [onfulfilled]
     * @param {(reason: any) => ERef<U>} [onrejected]
     * @returns {Promise<U>}
     * @readonly
     */
    readonly when: <T, U = T>(x: T | PromiseLike<T>, onfulfilled?: ((value: T) => ERef<U>) | undefined, onrejected?: ((reason: any) => ERef<U>) | undefined) => Promise<U>;
};
import type { HandledPromiseConstructor } from './types.js';
//# sourceMappingURL=E.d.ts.map